import os
import re
import time
from urllib.parse import urlparse
import threading
from typing import List, Tuple
from urllib.parse import unquote
from nsk.windows import FileUploader
import pandas as pd
from selenium.common.exceptions import (ElementClickInterceptedException,
                                        NoSuchElementException,
                                        StaleElementReferenceException,
                                        TimeoutException)
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC

from nsk.bot.base import IBot
from nsk.utils import retry_if_exception

class Microsoft365(IBot):
    def __init__(self, username: str, password: str, **kwargs):
        super().__init__(**kwargs)
        self.url = "https://login.microsoftonline.com/"
        self.authenticated = self._authentication(username, password)
        if not self.authenticated:
            raise ConnectionRefusedError("The username or password is incorrect.")
    
    @retry_if_exception(
        exceptions=(
            StaleElementReferenceException,
            ElementClickInterceptedException,
            TimeoutException,
        ),
        failure_return=False,
    )
    def _authentication(self, username: str, password: str) -> bool:
        self.navigate(self.url)
        time.sleep(self.retry_interval)
        if self.browser.current_url.startswith("https://m365.cloud.microsoft/?auth="):
            self.logger.info("Authenticated")
            return True
        try:
            # -- Username
            time.sleep(self.retry_interval)
            self.wait.until(
                EC.element_to_be_clickable((By.CSS_SELECTOR, 'input[type="email"]'))
            ).send_keys(username)
            # -- Next
            time.sleep(self.retry_interval)
            self.wait.until(
                EC.element_to_be_clickable((By.ID, "idSIButton9"))
            ).click()
            # -- Password
            time.sleep(self.retry_interval)
            self.wait.until(
                EC.element_to_be_clickable((By.CSS_SELECTOR, 'input[type="password"]'))
            ).send_keys(password)
            # -- Sign in
            time.sleep(self.retry_interval)
            self.wait.until(
                EC.element_to_be_clickable((By.CSS_SELECTOR, "input[id='idSIButton9']"))
            ).click()
        finally:
            time.sleep(self.retry_interval)
            try:
                # -- Alert
                alert = self.wait.until(
                    EC.presence_of_element_located((By.CSS_SELECTOR, "div[role='alert']"))
                )
                if alert.text:
                    raise ConnectionRefusedError(alert.text)
                return self._authentication(username, password)
            except TimeoutException:
                # -- Stay signed in
                self.wait.until(
                    EC.element_to_be_clickable((By.CSS_SELECTOR, "input[id='idSIButton9']"))
                ).click()
        time.sleep(self.retry_interval)
        if not self.browser.current_url.startswith("https://m365.cloud.microsoft/?auth="):
            raise ConnectionRefusedError("Authentication failed")
        self.logger.info("Authenticated to Microsoft 365")
        return True


class SharePoint(Microsoft365):
    redundant_name = r'[\ue000-\uf8ff]|\n|Press C to open file hover card'
    _upload_lock = threading.Lock()

    def __init__(self, url: str, username: str, password: str, **kwargs):
        super().__init__(
            username=username,
            password=password,
            **kwargs
        )
        self.url = f"{urlparse(url).scheme}://{urlparse(url).netloc}"
        self.navigate(self.url)
        if not self.browser.current_url.startswith(self.url) or self.browser.current_url.startswith("https://login.microsoftonline.com"):
            raise ConnectionRefusedError("Access to SharePoint site was denied.")

    def navigate(self, url, wait_for_complete = True):
        super().navigate(url, wait_for_complete)
        try:
            ms_error_header = self.wait.until(
                EC.element_to_be_clickable((By.CSS_SELECTOR,"div[id='ms-error-header']"))
            )
            self.logger.error(ms_error_header.text)
            self.wait.until(
                EC.element_to_be_clickable((By.CSS_SELECTOR,"div[id='ms-error'] a"))
            ).click()
            time.sleep(self.retry_interval)
            while self.browser.execute_script("return document.readyState") != "complete":
                continue
            time.sleep(self.retry_interval)
            return self.navigate(url, wait_for_complete)
        except Exception:
            pass
 
    @retry_if_exception(
        exceptions=(
            StaleElementReferenceException,
            ElementClickInterceptedException,
            TimeoutException,
        ),
        failure_return=pd.DataFrame()
    )
    def info(self,site_url:str) -> pd.DataFrame:
        """ Lấy danh sách file và folder từ 'site_url' trên SharePoint.
        Args:
            site_url (str): Đường dẫn URL của site SharePoint cần lấy thông tin.
        Returns:
            pd.DataFrame: DataFrame chứa thông tin các file và folder khớp với mẫu, 
        """
        self.logger.info(f"Get info {site_url}")
        result = []
        self.navigate(site_url)
        rows = self.wait.until(
            method = EC.any_of(
                EC.presence_of_all_elements_located((By.CSS_SELECTOR,"div[id^='virtualized-list'][role='row']")),
                EC.presence_of_all_elements_located((By.CSS_SELECTOR,"div[data-automationid='DetailsRow'][role='row']")),
            )
        )
        for row in rows:
            # -- Type
            type_el = row.find_element(
                by = By.CSS_SELECTOR,
                value="div[data-automationid='field-DocIcon'],div[data-automation-key^='fileTypeIconColumn']"
            )
            try:
                type_name = type_el.find_element(By.TAG_NAME,'img').get_attribute('alt')
            except NoSuchElementException:
                type_name = None
            # -- Name
            name_el = row.find_element(
                by = By.CSS_SELECTOR,
                value="div[data-automationid='field-LinkFilename'],div[data-automation-key^='displayNameColumn']"
            )
            name = name_el.text
            name = re.sub(self.__class__.redundant_name,'',name)
            # Agg
            result.append({
                "Type":type_name,
                "Name":name,
            })
        return pd.DataFrame(result)

    @retry_if_exception(
        exceptions=(
            StaleElementReferenceException,
            ElementClickInterceptedException,
            TimeoutException,
        ),
        failure_return=[(None,None,"Download Error"),]
    )
    def download(self,site_url:str,file_pattern:str) -> List[Tuple[str | None,str,str]]:
        self.logger.info(f"Search {file_pattern} - {site_url}")
        result = []
        self.navigate(site_url)
        # Folder
        folder_found = False
        folders: List[str] = file_pattern.split("/")[:-1]
        for step in folders:
            time.sleep(self.retry_interval)
            gridcells = self.wait.until(
                EC.any_of(
                    EC.presence_of_all_elements_located((By.CSS_SELECTOR,"div[role='gridcell'][data-automationid='field-LinkFilename']")),
                    EC.presence_of_all_elements_located((By.CSS_SELECTOR,"div[role='gridcell'][data-automation-key^='displayNameColumn']")),
                )
            )
            for gridcell in gridcells:
                text = re.sub(self.__class__.redundant_name,'',gridcell.text)
                if re.match(step,text):
                    button = gridcell.find_element(
                        By.XPATH,
                        './/button | .//span[@role="button"]'
                    )
                    self.wait.until(EC.element_to_be_clickable(button)).click()
                    time.sleep(self.timeout)
                    folder_found = True
                    break
            time.sleep(self.retry_interval)  
        if not folder_found:
            raise LookupError(f"Folder Not Found: {file_pattern}")
        # File
        file: str = file_pattern.split("/")[-1]
        gridcells = self.wait.until(
            EC.any_of(
                EC.presence_of_all_elements_located((By.CSS_SELECTOR,"div[role='gridcell'][data-automationid='field-LinkFilename']")),
                EC.presence_of_all_elements_located((By.CSS_SELECTOR,"div[role='gridcell'][data-automation-key^='displayNameColumn']")),
            )
        )
        for gridcell in gridcells:
            file_name = re.sub(self.__class__.redundant_name, '', gridcell.text)
            if re.match(file,file_name):
                button = gridcell.find_element(
                    By.XPATH,
                    './/button | .//span[@role="button"]'
                )
                self.wait.until(EC.element_to_be_clickable(button))
                time.sleep(self.retry_interval)
                # Copy Link
                link = None
                button.click()
                time.sleep(2)
                new_window = self.browser.window_handles[-1]  
                if new_window != self.root_window:
                    self.browser.switch_to.window(new_window)
                    while self.browser.execute_script("return document.readyState") != "complete":
                        time.sleep(self.retry_interval)
                    link = unquote(self.browser.current_url)
                    self.browser.close()
                    self.browser.switch_to.window(self.root_window)
                else:
                    close_button = self.wait.until(
                        EC.element_to_be_clickable((By.CSS_SELECTOR,"button[id='closeCommand']"))
                    )
                    time.sleep(self.retry_interval)
                    close_button.click()
                    time.sleep(self.timeout)
                # Download File
                ActionChains(self.browser).context_click(button).perform()
                download_btn = self.wait.until(
                    EC.presence_of_element_located(
                        (By.CSS_SELECTOR,"button[data-automationid='downloadCommand']")
                    )
                )
                self.wait.until(EC.element_to_be_clickable(download_btn)).click()
                time.sleep(5)
                file_path, status = self.wait_for_download_to_finish()
                self.logger.info(f"Download {file_name}: {file_path if not status else status},")
                result.append((link,os.path.join(self.download_directory,file_path),status))
        return result
        
    @retry_if_exception(
        exceptions=(
            StaleElementReferenceException,
            ElementClickInterceptedException,
            TimeoutException,
        ),
        failure_return=False,
    )    
    def upload(self,site_url:str, path: str) -> bool:
        with self._upload_lock:
            path = os.path.abspath(path)
            if not os.path.isfile(path):
                raise FileNotFoundError(f"File Not Found Error: {path}")
            self.logger.info(f"Upload to {unquote(site_url)}")
            # ----- Close -----
            time.sleep(self.retry_interval)
            while True:
                try:
                    uploader = FileUploader("Open")
                    uploader.close()
                except TimeoutException:
                    break
            # ----- Open -----
            self.navigate(site_url)
            time.sleep(self.retry_interval)
            self.wait.until(
                EC.element_to_be_clickable((By.CSS_SELECTOR,"button[data-id='UploadCommand']"))
            ).click()
            self.wait.until(
                EC.element_to_be_clickable((By.CSS_SELECTOR,f"button[data-automationid='uploadFileCommand']"))
            ).click()
            # --
            uploader = FileUploader("Open")
            if not uploader.upload(path):
                self.logger.error(f"Failed to upload {path}")
                return False
            time.sleep(1)
            message = self.wait.until(
                EC.presence_of_element_located((By.XPATH,"//div[@role='alert']//div[starts-with(@class, 'messageRow')]"))
            ).text
            if not message.startswith((
                f"Uploaded {os.path.basename(path)} to ",
                f"Uploading {os.path.basename(path)} to "
            )):
                self.logger.error(message)
                return False
            self.logger.info(message)
            return True
            
            
        
class PowerApp(Microsoft365):
    def __init__(self, url: str, username365:str,password365:str,username: str | None = None , password: str | None = None, **kwargs):
        """
        Args:
            url (str): PowerApp URL
            username365 (str): Microsoft 365 username
            password365 (str): Microsoft 365 password
            username (str | None, optional): PowerApp username. Defaults to None.
            password (str | None, optional): PowerApp password. Defaults to None.
        """
        super().__init__(
            username=username365,
            password=password365,
            **kwargs
        )
        self.url = url
        time.sleep(self.retry_interval)
        self.navigate(self.url)
        self.wait.until(EC.visibility_of_element_located((By.CSS_SELECTOR, "div[id='progressBarWrapper']")))
        self.wait.until(EC.invisibility_of_element_located((By.CSS_SELECTOR, "div[id='progressBarWrapper']")))
        self.wait.until(
            EC.frame_to_be_available_and_switch_to_it((By.CSS_SELECTOR,"iframe[id='fullscreen-app-host']"))
        )
        if username and password:
            self.wait.until(
                EC.element_to_be_clickable((By.CSS_SELECTOR,"input[appmagic-control='idtextbox']"))
            ).send_keys(username)
            self.wait.until(
                EC.element_to_be_clickable((By.CSS_SELECTOR,"input[appmagic-control='pwtextbox']"))
            ).send_keys(password)
            try:
                login_btn = self.wait.until(
                    EC.element_to_be_clickable((
                        By.XPATH,
                        "//button[contains(@class, 'appmagic-button-container') and contains(@class, 'no-focus-outline')]"
                        "[.//div[text()='Login']]"
                    ))
                )
                login_btn.click()
                self.wait.until(EC.invisibility_of_element(login_btn))
            except TimeoutException:
                raise ConnectionRefusedError("Authentication failed ")
            self.logger.info("Authenticated to PowerApp")


__all__ = ["Microsoft365","SharePoint","PowerApp"]



import os
import re
import time
from contextlib import suppress
from pathlib import Path
from typing import List, Literal

from pywinauto import Application
from pywinauto.application import WindowSpecification
from pywinauto.timings import wait_until_passes
from selenium.common.exceptions import (ElementClickInterceptedException,
                                        StaleElementReferenceException,
                                        TimeoutException)
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC

from nsk.bot.base import IBot
from nsk.utils import retry_if_exception


class MailDealer(IBot):
    def __init__(self, url: str, username: str, password: str, **kwargs):
        super().__init__(**kwargs)
        self.url = url
        self.authenticated = self._authentication(username, password)
        if not self.authenticated:
            raise ConnectionRefusedError("The username or password is incorrect.")
    @retry_if_exception(
        failure_return=False
    )
    def _authentication(self,username:str,password:str) -> bool:
        self.navigate(self.url)
        self.wait.until(
            method = EC.element_to_be_clickable((By.ID, "fUName")),
            message = "Not Found Username Field",
        ).send_keys(username)
        self.wait.until(
            method = EC.element_to_be_clickable((By.ID, "fPassword")),
            message="Not Found Password Field",
        ).send_keys(password)
        self.wait.until(
            method=EC.element_to_be_clickable((By.CSS_SELECTOR, "input[value='ログイン']")),
            message="Not Found Login Button",
        ).click()
        time.sleep(self.retry_interval)
        while self.browser.execute_script("return document.readyState") != "complete":
            time.sleep(self.retry_interval)
        with suppress(TimeoutException):
            error = self.wait.until(
                EC.presence_of_element_located(
                    (By.CSS_SELECTOR, "div[class='d_error_area ']"),
                ),
            )
            self.logger.error(error.text)
            return False
        with suppress(TimeoutException):
            olv_dialog = self.wait.until(
                EC.presence_of_element_located((By.CSS_SELECTOR,"div[class='olv-dialog']"))
            )
            time.sleep(self.retry_interval)
            for button in olv_dialog.find_elements(By.TAG_NAME,'button'):
                if button.text.find("同意する") != -1:
                    self.wait.until(
                        EC.element_to_be_clickable(button)
                    ).click()
                    break
        if self.browser.current_url.find("app") != -1:
            self.logger.info("Authenticated")
            return True
        return True

    @retry_if_exception(
        exceptions=(
            ElementClickInterceptedException,
            StaleElementReferenceException,
            TimeoutException,
        ),
        failure_return=False
    )
    def send_mail(self,fr:str,to:str,subject:str,content:str,attachments:List[str] = [],mode:Literal["送信確認","承認依頼確認","一時保存"] = "一時保存") -> bool:
        for window in self.browser.window_handles:
            if window != self.root_window:
                self.browser.switch_to.window(window)
                self.browser.close()
        self.browser.switch_to.window(self.root_window)
        self.logger.info(f"Send mail from {fr} to {to} (subject: {subject})")
        self.wait.until(
            method = EC.frame_to_be_available_and_switch_to_it((By.CSS_SELECTOR, "iframe[id='ifmSide']")),
            message = "Not Found Side Frame",
        )
        self.wait.until(
            method = EC.element_to_be_clickable((By.CSS_SELECTOR, "span[title='メール作成']")),
            message="Not Found Create Mail Button",
        ).click()
        while len(self.browser.window_handles) == 1:
            continue
        new_window = self.browser.window_handles[-1]
        self.browser.switch_to.window(new_window)
        self.browser.maximize_window()
        while self.browser.execute_script("return document.readyState") != "complete":
            time.sleep(self.retry_interval)
        self.wait.until(
            EC.element_to_be_clickable((By.XPATH, "//button[text()='次へ']"))
        ).click()
        # From
        fr_input = self.wait.until(
            EC.element_to_be_clickable((By.CSS_SELECTOR, "input[name='fFrom']"))
        )
        fr_input.clear()
        fr_input.send_keys(fr)
        # To
        to_input = self.wait.until(
            EC.element_to_be_clickable((By.CSS_SELECTOR, "input[name='fTo[]']"))
        )
        to_input.clear()
        to_input.send_keys(to)
        # Subject
        subject_input = self.wait.until(
            EC.element_to_be_clickable((By.CSS_SELECTOR,"input[id='fSubject']"))
        )
        subject_input.clear()
        subject_input.send_keys(subject)
        # Body
        self.wait.until(
            EC.frame_to_be_available_and_switch_to_it((By.TAG_NAME, "iframe"))
        )
        body = self.wait.until(EC.presence_of_element_located((By.TAG_NAME, "body")))
        self.browser.execute_script(
            "arguments[0].innerText = arguments[1];", body, content
        )
        self.browser.switch_to.default_content()
        for attachment in attachments:
            # Check
            if not os.path.exists(attachment):
                raise FileNotFoundError(f"Attachment Not Found: {attachment}")
            # Attachment
            self.wait.until(
                method = EC.element_to_be_clickable((By.CSS_SELECTOR, "button[title='添付ファイル']")),
                message="Not Found Attachment Button",
            ).click()
            file_uploader: WindowSpecification = wait_until_passes(
                timeout=self.timeout,
                retry_interval=self.retry_interval,
                func=lambda: Application()
                .connect(title="Open")
                .window(title="Open"),
            )
            wait_until_passes(
                timeout=self.timeout,
                retry_interval=self.retry_interval,
                func=lambda dialog=file_uploader: dialog 
                .child_window(class_name="ComboBoxEx32")
                .child_window(class_name="Edit")
                .wrapper_object(),
            ).type_keys(re.sub(r"([{}^%~()])", r"{\1}", str(Path(attachment))),with_spaces=True)
            wait_until_passes(
                timeout=self.timeout,
                retry_interval=1,
                func=lambda dialog=file_uploader: dialog.child_window(
                    title="&Open", class_name="Button"
                ).wrapper_object(),
            ).click_input()
            for button in self.wait.until(
                method = EC.presence_of_all_elements_located((By.CSS_SELECTOR,"button[class='accent-btn__btn']")),
                message="Not Found Send Button"
            ):
                if button.text.find(mode) != -1:
                    self.browser.execute_script(
                        "arguments[0].scrollIntoView({behavior: 'smooth', block: 'center'});",
                        button,
                    )
                    time.sleep(1)
                    self.wait.until(EC.element_to_be_clickable(button)).click()
                    time.sleep(1)
                    snackbar = self.wait.until(
                        method = EC.presence_of_element_located((By.CSS_SELECTOR, "div[class='snackbar__msg']")),
                        message= "Snackbar Not Found ", 
                    )
                    self.logger.info(snackbar.text)
                    break
        self.browser.close()
        self.browser.switch_to.window(self.root_window)
        self.logger.info("Send mail successfully")
        return True



from .dandoli import Dandoli
from .mail_dealer import MailDealer
from .microsoft import Microsoft365,SharePoint,PowerApp
from .web_access import WebAccess

__all__ = ['Microsoft365','SharePoint','PowerApp','WebAccess','MailDealer','Dandoli']



import time
from contextlib import suppress
from typing import List
from urllib.parse import urljoin

from selenium import webdriver
from selenium.common.exceptions import (ElementClickInterceptedException,
                                        StaleElementReferenceException,
                                        TimeoutException)
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support import expected_conditions as EC

from nsk.bot.base import IBot
from nsk.utils import retry_if_exception


class Dandoli(IBot):
    def __init__(
        self,
        url: str,
        username: str,
        password: str,
        **kwargs,
    ):
        if 'options' not in kwargs:
            kwargs['options'] = webdriver.ChromeOptions()
        prefs = kwargs['options'].experimental_options.get("prefs", {})
        prefs["profile.password_manager_leak_detection"] = False # Disable 'Change the password'
        kwargs['options'].add_experimental_option("prefs", prefs)
        if 'timeout' not in kwargs or kwargs['timeout'] < 10:
            kwargs['timeout'] = 10 # Min Timeout is 10s
        # ----- Init -----
        super().__init__(**kwargs)
        self.url = url
        self.authenticated = self._authentication(username, password)
        if not self.authenticated:
            raise ConnectionRefusedError("The username or password is incorrect.")
    
    def navigate(self, url, wait_for_complete = True):
        super().navigate(url, wait_for_complete)
        with suppress(TimeoutException):
            self.wait.until(
                EC.visibility_of_element_located((By.CSS_SELECTOR, "div[class='loader-container']"))
            )
            self.wait.until(
                EC.invisibility_of_element_located((By.CSS_SELECTOR, "div[class='loader-container']"))
            )
        
    @retry_if_exception(
        exceptions=(
            StaleElementReferenceException,
            ElementClickInterceptedException,
        ),
        failure_return=False,
    )
    def _authentication(self, username: str, password: str) -> bool:
        self.navigate(self.url)
        self.wait.until(
            EC.element_to_be_clickable((By.CSS_SELECTOR, "input[name='username']"))
        ).send_keys(username)
        self.wait.until(
            EC.element_to_be_clickable((By.CSS_SELECTOR, "input[id='password']"))
        ).send_keys(password)
        self.wait.until(
            EC.element_to_be_clickable((By.CSS_SELECTOR, "button[class='btn login-btn']"))
        ).click()
        time.sleep(self.retry_interval)
        while self.browser.execute_script("return document.readyState") != "complete":
            time.sleep(self.retry_interval)
        with suppress(TimeoutException):
            message = self.wait.until(
                EC.presence_of_element_located((By.CSS_SELECTOR, "div[id='field-message']"))
            )
            self.logger.error(message.text.replace("\n", ""))
            return False
        while self.browser.execute_script("return document.readyState") != "complete":
            time.sleep(self.retry_interval)
        self.wait.until(
            method = EC.element_to_be_clickable((By.CSS_SELECTOR, "div[id='header-account']")),
            message="Not Found Header Account",
        )
        self.logger.info("Authenticated")
        return True
    

    @retry_if_exception(
        exceptions=(
            StaleElementReferenceException,
            ElementClickInterceptedException,
        ),
        failure_return=False,
    )
    def _switch_place(self,name:str,company:str = "") -> bool:
        self.navigate(self.url)
        while True:
            with suppress(TimeoutException):
                self.wait.until(
                    method = EC.element_to_be_clickable((By.CSS_SELECTOR, "span[class='placeSwitchButton']")),
                    message="Not Found SwitchButton",
                ).click()
                time.sleep(self.retry_interval)
                self.wait.until(
                    EC.visibility_of_element_located((By.CSS_SELECTOR, "div.popover.bottom.in[role='tooltip'][id^='popover']"))
                )
                break
        time.sleep(self.retry_interval)
        self.wait.until(
            EC.visibility_of_element_located((By.CSS_SELECTOR, "div.popover.bottom.in[role='tooltip'][id^='popover']"))
        )
        time.sleep(self.retry_interval)
        xpath = (
            f"//a[contains(@class, 'popover__switchablePlaceListLink js-switch-place') and "
            f".//div[contains(@class, 'popover__switchablePlaceListName') and contains(., '{name}')] and "
            f".//div[contains(@class, 'popover__switchableCompanyName') and contains(., '{company}')]]"
        )
        places:List[WebElement]  = self.wait.until(
            EC.presence_of_all_elements_located((By.XPATH, xpath))
        )
        if len(places) != 1:
            raise RuntimeError(f"Expected exactly 1 place {name} of company '{company}', but found {len(places)}: {places}")
        # -- Redirect --
        self.browser.execute_script(
            "arguments[0].scrollIntoView({behavior: 'smooth', block: 'center'});",
            places[0]
        )
        self.logger.info("Switch place {name}, {company}".format(
            name = places[0].find_element(By.CSS_SELECTOR,"div[class='popover__switchablePlaceListName']").text.split("\n")[0].strip(),
            company = places[0].find_element(By.CSS_SELECTOR,"div[class='popover__switchableCompanyName']").text.strip()
        ))
        time.sleep(self.retry_interval)
        switching_btn = places[0].find_element(By.TAG_NAME,'span')
        self.wait.until(EC.element_to_be_clickable(switching_btn)).click()
        with suppress(TimeoutException):
            self.wait.until(
                EC.visibility_of_element_located((By.CSS_SELECTOR, "div[class='loader-container']"))
            )
            self.wait.until(
                EC.invisibility_of_element_located((By.CSS_SELECTOR, "div[class='loader-container']"))
            )
        time.sleep(self.retry_interval)
        # -- Check Place --
        current_place = self.wait.until(
            EC.visibility_of_element_located((By.CSS_SELECTOR, "div[class='placeSwitchButton__placeName']"))
        )
        return name in current_place.text
    
    @retry_if_exception(
        exceptions=(
          StaleElementReferenceException,
          ElementClickInterceptedException,  
        ),
        failure_return=[]
    )
    def find_place(self,place:str) -> List[dict[str,str]]:
        places = []
        self.navigate(self.url)
        while True:
            with suppress(TimeoutException):
                self.wait.until(
                    method = EC.element_to_be_clickable((By.CSS_SELECTOR, "span[class='placeSwitchButton']")),
                    message="Not Found SwitchButton",
                ).click()
                time.sleep(self.retry_interval)
                self.wait.until(
                    EC.visibility_of_element_located((By.CSS_SELECTOR, "div.popover.bottom.in[role='tooltip'][id^='popover']"))
                )
                break
        time.sleep(self.retry_interval)
        self.wait.until(
            EC.visibility_of_element_located((By.CSS_SELECTOR, "div.popover.bottom.in[role='tooltip'][id^='popover']"))
        )
        time.sleep(self.retry_interval)
        found:List[WebElement]  = self.wait.until(
            EC.presence_of_all_elements_located((By.XPATH, (
                f"//a[contains(@class, 'popover__switchablePlaceListLink js-switch-place') and "
                f".//div[contains(@class, 'popover__switchablePlaceListName') and contains(., '{place}')]]"
            )))
        )
        for p in found:
            places.append({
                'name': p.find_element(By.CSS_SELECTOR,"div[class='popover__switchablePlaceListName']").text.split("\n")[0].strip(),
                'company': p.find_element(By.CSS_SELECTOR,"div[class='popover__switchableCompanyName']").text.strip(),
            })
        return places

    @retry_if_exception(
        exceptions=(
            StaleElementReferenceException,
            ElementClickInterceptedException,
        ),
        failure_return={},
    )
    def get_building_info(self,name:str,place:str,company:str="", keywords: list[str] = []) -> dict[str,str | None]:
        """
        Retrieves information about a construction/building project based on the provided details.

        Args:
            name (str): Name of the building or project.
            place (str): Location of the building or project.
            company (str, optional): Name of the associated company. Defaults to an empty string.
            keywords (list[str], optional): Additional keywords to refine the search or filter the results. Defaults to an empty list.

        Returns:
            dict[str, str]: A dictionary containing relevant building information, such as address, contractor, status, etc.
        """
        self.logger.info(
            f"Getting {keywords} from building '{name}' located in '{place}' (company: '{company}')"
        )
        result: dict[str,str | None] = {}
        if not self._switch_place(
            name=place,
            company=company
        ):
            if sites := self.find_place(place):
                for site in sites:
                    if result := self.get_building_info(
                        name=name,
                        place=site['name'],
                        company=site['company'],
                        keywords=keywords
                    ):
                        return result
            raise RuntimeError(f"Place {place} not found.")
        nav_site = self.wait.until(
            EC.element_to_be_clickable((By.CSS_SELECTOR, "a[id='nav-sites']"))
        )
        redirect_url=urljoin(self.url,nav_site.get_attribute('href'))
        self.navigate(redirect_url)
        self.logger.info(f"Search {name}")
        self.wait.until(
            EC.element_to_be_clickable((By.CSS_SELECTOR, "input[name='keyword'][type='text']"))
        ).send_keys(name)
        self.wait.until(
            EC.element_to_be_clickable((By.XPATH, "//button[@type='submit' and text()='検索']"))
        ).click()
        while True:
            request_message = self.wait.until(
                method= EC.presence_of_element_located(
                    (By.CSS_SELECTOR, "div[class='pageContentSite__requestingMessage js-requesting-message']")
                ),
                message="Not Found requestMessage"
            )
            if request_message.get_attribute("style") == "display: none;":
                break
            time.sleep(self.retry_interval)
        rows = self.wait.until(
            EC.presence_of_all_elements_located((
                By.XPATH,
                '//div[contains(@class, "pageContentSite__tableBodyWrap")]//tr'
            ))
        )
        if len(rows) != 1:
            raise RuntimeError("Found more than one building.")
        if rows[0].get_attribute("class") == "empty":
            raise RuntimeError(f"The building {name} is not found.")
        self.wait.until(
            EC.element_to_be_clickable(rows[0])
        ).click()
        with suppress(TimeoutException):
            self.wait.until(
                EC.visibility_of_element_located((By.CSS_SELECTOR, "div[class='loader-container']"))
            )
            self.wait.until(
                EC.invisibility_of_element_located((By.CSS_SELECTOR, "div[class='loader-container']"))
            )
        for keyword in keywords:
            try:
                value = self.wait.until(
                    EC.visibility_of_element_located((By.XPATH, f"//th[contains(normalize-space(text()), '{keyword}')]/following-sibling::td[1]"))
                )
                self.browser.execute_script(
                    "arguments[0].scrollIntoView({behavior: 'smooth', block: 'center'});",
                    value
                )
                time.sleep(self.retry_interval)
                value = value.text.strip()
            except TimeoutException:
                value = None
            self.logger.info(f"Found {value} for {keyword}")
            result[keyword] = value
        return result
            
        


import time
import logging
import win32api
import win32gui
import win32con
from typing import List,Tuple
from selenium.common.exceptions import TimeoutException

class Windows:
    def __init__(
        self, 
        class_name: str | None = None, 
        window_name: str | None = None, 
        timeout: float = 5.0, 
        retry_interval: float = 0.5,
        log_name: str | None = None,
    ):
        self.class_name = class_name
        self.window_name = window_name
        self.timeout = timeout
        self.retry_interval = retry_interval
        self.hwnd = self._init_window()
        self.logger = logging.getLogger(log_name)
    
    @classmethod
    def from_hwnd(cls, hwnd: int, log_name: str | None = None):
        if not hwnd or not win32gui.IsWindow(hwnd):
            raise ValueError(f"Invalid Window Handle: {hwnd}")
        # --
        instance = cls.__new__(cls)
        instance.hwnd = hwnd
        instance.class_name = None
        instance.window_name = None
        instance.timeout = 5.0
        instance.retry_interval = 0.5
        instance.logger = logging.getLogger(log_name)
        try:
            instance.class_name = win32gui.GetClassName(hwnd)
            instance.window_name = win32gui.GetWindowText(hwnd)
        except Exception as e:
            instance.logger.warning(f"Could not retrieve window info: {e}")
        return instance
        

    def _init_window(
        self, 
    ) -> int:
        start_time = time.time()
        while True:
            if hwnd := win32gui.FindWindow(self.class_name, self.window_name):
                return hwnd
            if (time.time() - start_time) > self.timeout:
                raise TimeoutException("Window not found within the timeout period.")
            time.sleep(self.retry_interval)

    def enum_child_windows(self) -> List[Tuple[int, str, str]]:
        """Liệt kê tất cả child windows và controls"""
        def callback(hwnd: int, results: list):
            class_name = win32gui.GetClassName(hwnd)
            window_text = win32gui.GetWindowText(hwnd)
            results.append((hwnd, class_name, window_text))
            return True
        time.sleep(self.retry_interval)
        child_windows = []
        win32gui.EnumChildWindows(self.hwnd, callback, child_windows)
        return child_windows
    
    def find_child(self, class_name: str = None, text: str = None):
        time.sleep(self.retry_interval)
        children = self.enum_child_windows()
        for hwnd, child_class, child_text in children:
            if class_name and class_name != child_class:
                continue
            if text and text != child_text:
                continue
            return self.__class__.from_hwnd(hwnd)
        
    def find_children(self, class_name: str = None, text: str = None) -> List['Windows']:
        time.sleep(self.retry_interval)
        windows = []
        children = self.enum_child_windows()
        for hwnd, child_class, child_text in children:
            if class_name and class_name != child_class:
                continue
            if text and text != child_text:
                continue
            windows.append(self.__class__.from_hwnd(hwnd))
        return windows
    
    def set_text(
        self,
        message:str,
        wparam:int = 0,
    ) -> int:    
        return win32gui.SendMessage(
            self.hwnd,
            win32con.WM_SETTEXT,    
            wparam,
            message,
        )

    def click(self, double_click: bool = False): 
        win32gui.SetForegroundWindow(self.hwnd)
        x_screen, y_screen = win32gui.ClientToScreen(self.hwnd, (0, 0))
        win32api.SetCursorPos((x_screen, y_screen))
        count = 2 if double_click else 1
        for _ in range(count):
            win32gui.PostMessage(self.hwnd, win32con.BM_CLICK, 0, 0)
            time.sleep(0.05)
        

    def minimize(self):
        """Minimize window"""
        win32gui.ShowWindow(self.hwnd, win32con.SW_MINIMIZE)

    def maximize(self):
        """Maximize window"""
        win32gui.ShowWindow(self.hwnd, win32con.SW_MAXIMIZE)

    def restore(self):
        """Restore window to its previous size"""
        win32gui.ShowWindow(self.hwnd, win32con.SW_RESTORE)

    def close(self):
        """Close window"""
        win32gui.PostMessage(self.hwnd, win32con.WM_CLOSE, 0, 0)
    
__all__ = ['Windows']


import os
import time
import win32gui
from nsk.windows.base import Windows
from pywintypes import error
from nsk.utils.decorator import retry_if_exception

class FileUploader(Windows):
    def __init__(self,window_name = None, timeout: float = 5, retry_interval: float = 0.5):
        super().__init__("#32770", window_name, timeout, retry_interval)
        
    @retry_if_exception(
        exceptions=(
            error
        ),
        failure_return=False
    )
    def upload(self,path: str) -> bool:
        if not os.path.isfile(path):
            raise FileNotFoundError(f"File Not Found Error: {path} ")
        edit = self.find_child(class_name="Edit")
        edit.set_text(path)
        upload = self.find_child(class_name="Button",text="&Open")
        upload.click()
        timeout = time.time() + self.timeout
        while time.time() < timeout:
            if not win32gui.IsWindow(self.hwnd):
                return True 
            time.sleep(self.retry_interval)
        return False